"""
Calculations app configuration
"""

from django.apps import AppConfig


class CalculationsConfig(AppConfig):
    """
    Calculations configuration
    """
    name = "djangophysics.calculations"
    verbose_name = "Calculations"
