# Generated by Django 3.1.5 on 2021-04-19 15:48

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('units', '0004_auto_20210418_1916'),
    ]

    operations = [
        migrations.AlterField(
            model_name='customunit',
            name='id',
            field=models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
    ]
